from __future__ import annotations

from dataclasses import dataclass

from .profiler import ProfileNode, RequestTrace


@dataclass(slots=True)
class FlatNode:
    node: ProfileNode
    start_ms: float
    duration_ms: float
    percent_total: float


def _flatten(
    node: ProfileNode,
    root_started_ns: int,
    total_ns: int,
) -> list[FlatNode]:
    flat: list[FlatNode] = []
    duration_ms = node.duration_ns / 1_000_000
    start_ms = (node.started_ns - root_started_ns) / 1_000_000
    percent_total = (
        0.0 if total_ns == 0 else (node.duration_ns / total_ns) * 100
    )
    flat.append(
        FlatNode(
            node=node,
            start_ms=start_ms,
            duration_ms=duration_ms,
            percent_total=percent_total,
        )
    )
    for child in node.children:
        flat.extend(_flatten(child, root_started_ns, total_ns))
    return flat


def _render_bar(
    start_ms: float,
    duration_ms: float,
    total_ms: float,
    width: int,
) -> str:
    if width <= 0:
        return ""

    canvas = [" "] * width
    if total_ms <= 0:
        return "".join(canvas)

    start_col = min(width - 1, int((start_ms / total_ms) * width))
    span_col = max(1, int((duration_ms / total_ms) * width))
    end_col = min(width, start_col + span_col)
    for index in range(start_col, end_col):
        canvas[index] = "#"
    return "".join(canvas)


def _format_label(node: ProfileNode) -> str:
    if node.filename == "<request>":
        return node.name
    return f"{node.filename}:{node.lineno} {node.name}"


def render_trace(
    trace: RequestTrace,
    *,
    width: int = 60,
    hotspot_count: int = 5,
) -> str:
    total_ns = trace.duration_ns
    total_ms = total_ns / 1_000_000
    flat_nodes = _flatten(trace.root, trace.root.started_ns, total_ns)

    non_root_nodes = flat_nodes[1:]
    hottest_node = max(
        non_root_nodes,
        key=lambda item: item.node.self_time_ns,
        default=None,
    )
    hottest_self = sorted(
        non_root_nodes,
        key=lambda item: item.node.self_time_ns,
        reverse=True,
    )[:hotspot_count]

    lines = [
        (
            f"Request {trace.status_code} {trace.method} "
            f"{trace.path} took {total_ms:.2f} ms"
        ),
        "Hotspots",
    ]

    if hottest_self:
        for item in hottest_self:
            lines.append(
                "  "
                f"{item.duration_ms:7.2f} ms total | "
                f"{item.node.self_time_ns / 1_000_000:7.2f} ms self  "
                f"{_format_label(item.node)}"
            )
    else:
        lines.append("  no Python frames were captured")

    lines.append("Waterfall")
    for item in flat_nodes:
        marker = ""
        if hottest_node is not None and item.node is hottest_node.node:
            marker = "  <<< hottest"

        bar = _render_bar(item.start_ms, item.duration_ms, total_ms, width)
        lines.append(
            f"{item.start_ms:8.2f} ms |{bar}|"
            f" {item.duration_ms:8.2f} ms"
            f" {item.percent_total:5.1f}% {_format_label(item.node)}{marker}"
        )

    if trace.dropped_calls:
        lines.append(f"Dropped calls: {trace.dropped_calls}")

    return "\n".join(lines)



from __future__ import annotations

import asyncio
import inspect
import os
import sys
import threading
import time
from contextvars import ContextVar, Token
from dataclasses import dataclass, field
from types import FrameType


_ACTIVE_RECORDER: ContextVar["RequestProfiler | None"] = ContextVar(
    "waterfall_active_recorder",
    default=None,
)
_THREAD_PROFILE_STATE = threading.local()
_PACKAGE_ROOT = os.path.dirname(os.path.abspath(__file__))


@dataclass(slots=True)
class ProfileNode:
    name: str
    filename: str
    lineno: int
    started_ns: int
    ended_ns: int | None = None
    depth: int = 0
    children: list["ProfileNode"] = field(default_factory=list)

    @property
    def duration_ns(self) -> int:
        if self.ended_ns is None:
            return 0
        return max(0, self.ended_ns - self.started_ns)

    @property
    def self_time_ns(self) -> int:
        child_time = sum(child.duration_ns for child in self.children)
        return max(0, self.duration_ns - child_time)


@dataclass(slots=True)
class RequestTrace:
    method: str
    path: str
    status_code: int
    root: ProfileNode
    dropped_calls: int = 0

    @property
    def duration_ns(self) -> int:
        return self.root.duration_ns


def _current_task() -> asyncio.Task[object] | None:
    try:
        return asyncio.current_task()
    except RuntimeError:
        return None


def _install_dispatcher() -> None:
    depth = getattr(_THREAD_PROFILE_STATE, "depth", 0)
    if depth == 0:
        _THREAD_PROFILE_STATE.previous_profile = sys.getprofile()
        sys.setprofile(_dispatch_profile_event)
    _THREAD_PROFILE_STATE.depth = depth + 1


def _uninstall_dispatcher() -> None:
    depth = getattr(_THREAD_PROFILE_STATE, "depth", 0)
    if depth <= 1:
        sys.setprofile(
            getattr(_THREAD_PROFILE_STATE, "previous_profile", None)
        )
        _THREAD_PROFILE_STATE.depth = 0
        _THREAD_PROFILE_STATE.previous_profile = None
        return
    _THREAD_PROFILE_STATE.depth = depth - 1


def _dispatch_profile_event(frame: FrameType, event: str, arg: object) -> None:
    recorder = _ACTIVE_RECORDER.get()
    if recorder is not None:
        recorder.handle_event(frame, event, arg)

    previous_profile = getattr(_THREAD_PROFILE_STATE, "previous_profile", None)
    if previous_profile is not None:
        previous_profile(frame, event, arg)


def _format_frame_name(frame: FrameType) -> tuple[str, str, int]:
    code = frame.f_code
    filename = os.path.relpath(code.co_filename, os.getcwd())
    function_name = (
        code.co_qualname if hasattr(code, "co_qualname") else code.co_name
    )
    return function_name, filename, code.co_firstlineno


class RequestProfiler:
    def __init__(
        self,
        method: str,
        path: str,
        *,
        max_nodes: int = 5000,
    ) -> None:
        started_ns = time.perf_counter_ns()
        self.trace = RequestTrace(
            method=method,
            path=path,
            status_code=500,
            root=ProfileNode(
                name=f"{method} {path}",
                filename="<request>",
                lineno=0,
                started_ns=started_ns,
                depth=0,
            ),
        )
        self.max_nodes = max_nodes
        self.task: asyncio.Task[object] | None = None
        self._frame_nodes: dict[int, ProfileNode] = {}
        self._token: Token[RequestProfiler | None] | None = None

    def start(self) -> None:
        self.task = _current_task()
        self.trace.root.started_ns = time.perf_counter_ns()
        self._token = _ACTIVE_RECORDER.set(self)
        _install_dispatcher()

    def stop(self, *, status_code: int) -> RequestTrace:
        ended_ns = time.perf_counter_ns()
        self.trace.status_code = status_code

        try:
            if self._token is not None:
                _ACTIVE_RECORDER.reset(self._token)
        finally:
            _uninstall_dispatcher()

        for node in self._frame_nodes.values():
            if node.ended_ns is None:
                node.ended_ns = ended_ns
        self._frame_nodes.clear()
        self.trace.root.ended_ns = ended_ns
        return self.trace

    def handle_event(self, frame: FrameType, event: str, arg: object) -> None:
        if event not in {"call", "return"}:
            return

        if not self._is_current_request_task():
            return

        if event == "call":
            self._on_call(frame)
            return

        self._on_return(frame, arg)

    def _is_current_request_task(self) -> bool:
        if self.task is None:
            return True
        return _current_task() is self.task

    def _on_call(self, frame: FrameType) -> None:
        if not self._should_capture(frame):
            return

        if id(frame) in self._frame_nodes:
            return

        if len(self._frame_nodes) >= self.max_nodes:
            self.trace.dropped_calls += 1
            return

        parent = self._find_parent(frame.f_back)
        name, filename, lineno = _format_frame_name(frame)
        node = ProfileNode(
            name=name,
            filename=filename,
            lineno=lineno,
            started_ns=time.perf_counter_ns(),
            depth=parent.depth + 1,
        )
        parent.children.append(node)
        self._frame_nodes[id(frame)] = node

    def _on_return(self, frame: FrameType, arg: object) -> None:
        if self._is_coroutine_suspension(frame, arg):
            return

        node = self._frame_nodes.pop(id(frame), None)
        if node is not None and node.ended_ns is None:
            node.ended_ns = time.perf_counter_ns()

    def _find_parent(self, frame: FrameType | None) -> ProfileNode:
        while frame is not None:
            node = self._frame_nodes.get(id(frame))
            if node is not None:
                return node
            frame = frame.f_back
        return self.trace.root

    def _should_capture(self, frame: FrameType) -> bool:
        filename = os.path.abspath(frame.f_code.co_filename)
        return not filename.startswith(_PACKAGE_ROOT)

    def _is_coroutine_suspension(
        self,
        frame: FrameType,
        arg: object,
    ) -> bool:
        if not frame.f_code.co_flags & inspect.CO_COROUTINE:
            return False

        return asyncio.isfuture(arg) and not arg.done()



from __future__ import annotations

import sys
from typing import TextIO

from .profiler import RequestProfiler
from .render import render_trace


class WaterfallMiddleware:
    def __init__(
        self,
        app,
        *,
        stream: TextIO | None = None,
        width: int = 60,
        hotspot_count: int = 5,
        max_nodes: int = 5000,
    ) -> None:
        self.app = app
        self.stream = stream if stream is not None else sys.stdout
        self.width = width
        self.hotspot_count = hotspot_count
        self.max_nodes = max_nodes

    async def __call__(self, scope, receive, send) -> None:
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        method = scope.get("method", "HTTP")
        path = scope.get("path", "/")
        status_code = 500
        profiler = RequestProfiler(method, path, max_nodes=self.max_nodes)

        async def send_wrapper(message) -> None:
            nonlocal status_code
            if message["type"] == "http.response.start":
                status_code = message["status"]
            await send(message)

        profiler.start()
        try:
            await self.app(scope, receive, send_wrapper)
        finally:
            trace = profiler.stop(status_code=status_code)
            self.stream.write(
                render_trace(
                    trace,
                    width=self.width,
                    hotspot_count=self.hotspot_count,
                )
            )
            self.stream.write("\n")
            self.stream.flush()
